# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-11-30"
# Created: 2017-11-27 21:16

import abc


def format_vars(instance):
    attrs = vars(instance)
    return ", ".join("{}={}".format(key, value) for key, value in attrs.items())


class FromToDictBase(object):
    __metaclass__ = abc.ABCMeta

    @classmethod
    def from_dict(cls, d):
        new = cls()
        if not d:
            return new
        attrs = vars(new)
        for key in d:
            if key in attrs:
                # both in dict and this class
                setattr(new, key, d[key])
        return new

    def to_dict(self):
        attrs = vars(self)
        res = {}
        for key, value in attrs.items():
            if isinstance(value, FromToDictBase):
                res[key] = value.to_dict()
            else:
                res[key] = value
        return res


class PrintableBase(object):
    __metaclass__ = abc.ABCMeta

    def __str__(self):
        return "<{}>({})".format(self.__class__.__name__, format_vars(self))

    def __unicode__(self):
        return self.__str__()

    def __repr__(self):
        return self.__str__()


class Entry(PrintableBase, FromToDictBase):

    def __init__(self, uuid=None, created=None, updated=None):
        super(Entry, self).__init__()
        self.uuid = uuid
        """ :type : str | unicode """
        self.created = created
        """ :type : datetime.datetime """
        self.updated = updated
        """ :type : datetime.datetime """


class Manga(Entry):

    def __init__(self, uuid=None, name=None, chapter=None):
        super(Manga, self).__init__(uuid)
        self.name = name
        """ :type : str | unicode """
        self.chapter = chapter
        """ Latest chapter read by user
            :type : float """
        self.latest_chapter = 0
        """ Latest chapter found anywhere
            :type : int """
        self.urls = []
        """ :type : list[str | unicode] """


class User(Entry):
    ROLE_ADMIN = 0
    ROLE_USER = 1

    def __init__(self, uuid=None, firstname=None, lastname=None):
        super(User, self).__init__(uuid)
        self.firstname = firstname
        """ :type : str | unicode """
        self.lastname = lastname
        """ :type : str | unicode """
        self.role = User.ROLE_USER



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.5.4"
__date__ = "2018-12-01"
# Created: 2017-11-27 21:08

from .controller import IDFXManga
from .model import Manga, User

__all__ = ["IDFXManga", "Manga", "User"]



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__copyright__ = "Copyright (C) 2015-17, Florian JUNG"
__license__ = "All rights reserved"
__version__ = "0.1.1"
__date__ = "2017-11-27"
# Created: 2015-03-13 12:13

from abc import ABCMeta, abstractmethod
import datetime

from flotils.loadable import Loadable


class DAOException(Exception):
    pass


class AbstractMangaDao(Loadable):
    __metaclass__ = ABCMeta

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        super(AbstractMangaDao, self).__init__(settings)

    @abstractmethod
    def connect(self):
        """
        Connect to datasource

        :rtype: None
        :raises DAOException: On failure
        """
        raise NotImplementedError("Please implement")

    @abstractmethod
    def disconnect(self):
        """
        Disconnect from datasource

        :rtype: None
        :raises DAOException: On failure
        """
        raise NotImplementedError("Please implement")

    @abstractmethod
    def create_location(self, location):
        """
        Create the location if not exists

        :param location: save location
        :type location: str | unicode
        :rtype: None
        :raises DAOException: On failure
        """
        raise NotImplementedError("Please implement")

    @abstractmethod
    def create(self, manga):
        """
        Create a new manga

        :param manga: Manga to create
        type manga: idefix.model.Manga
        :return: Response (TODO: standardize)
        :rtype: None
        :raises DAOException: On failure
        """
        raise NotImplementedError("Please implement")

    @abstractmethod
    def update(self, manga):
        """
        Update a manga

        :param manga: Manga to update
        type manga: idefix.model.Manga
        :return: Response (TODO: standardize)
        :rtype: None
        :raises DAOException: On failure
        """
        raise NotImplementedError("Please implement")

    @abstractmethod
    def delete(self, manga):
        """
        Delete a manga

        :param manga: Manga to delete
        type manga: idefix.model.Manga
        :return: Response (TODO: standardize)
        :rtype: None
        :raises DAOException: On failure
        """
        raise NotImplementedError("Please implement")

    @staticmethod
    def format_dict(val, manga):
        if not manga:
            #print "no post"
            return val
        if not isinstance(val, dict):
            #print "not dict"
            return val
        if "_manga" not in val:
            #print "no _post"
            return val
        res = getattr(manga, val['_post'])
        if "+" in val:
            append = val['+']
            if isinstance(res, datetime.datetime):
                res = res.date()
            if isinstance(res, datetime.date):
                res = res + datetime.timedelta(**append)
            else:
                res += append
        if "-" in val:
            append = val['-']
            if isinstance(res, datetime.datetime):
                res = res.date()
            if isinstance(res, datetime.date):
                res = res - datetime.timedelta(**append)
            else:
                res -= append
        return res

    @abstractmethod
    def query(self, query, manga=None):
        """
        Query the datasource

        :param query: Query to apply
        location: str
            location in the database - None | "" for all (default: None)
        query: dict
            field_name: dict
                < | <=: value   smaller than
                > | >=: value   greater than
                ==: value       equals to
                range: dict
                    <: value    upper bound
                    >: value    lower bound
        sort_by: list[dict]
            field_name: asc | desc
        :type query: dict
        :param manga: If present insert into query (default: None)
        :type manga: idefix.model.Manga | None
        :return: Resulting list of mangas
        rtype: collections.Iterable[rss.post.Post]
        :raises DAOException: On failure
        """
        raise NotImplementedError("Please implement")

    @abstractmethod
    def flush(self, location=None):
        """
        Persist changes to datasource

        :param location: Which location to flush - None | "" for all
                (default: None)
        :type location: str | unicode
        :return: Response (TODO: standardize)
        :rtype: None
        :raises DAOException: On failure
        """
        raise NotImplementedError("Please implement")



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-12-01"
# Created: 2017-12-01 19:26

from flotils.loadable import load_file

from .model import Manga


def read_manga_file(path):
    """
    Read mangas from old manga file

    :param path: Read from file
    :type path: str | unicode
    :return: Read mangas
    :rtype: idefix.model.Manga
    """
    res = []

    loaded = load_file(path)

    for l in loaded.values():
        res.append(Manga(name=l['name'], chapter=l['chapter']))

    return res



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-12-01"
# Created: 2017-12-01 21:15

import logging
import logging.config
import argparse
import datetime

from flotils import get_logger
from flotils.logable import default_logging_config
from requests.compat import urlparse

from .controller import IDFXManga
from .model import Manga


logger = get_logger()
logging.captureWarnings(True)


def setup_parser():
    """
    Create and init argument parser

    :return: Argument parser
    :rtype: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(prog="idefix")
    parser.add_argument(
        "--debug", action="store_true",
        help="Use debug level output"
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + __version__
    )
    parser.add_argument(
        "-s", "--settings", nargs="?",
        help="Settings file"
    )
    parser.add_argument(
        "--pre_path", nargs="?", default=None,
        help="Base path to use"
    )
    parser.add_argument(
        "--manga_file", nargs="?", default=None,
        help="Manga file to use"
    )
    parser.add_argument(
        "-c", "--check", nargs="?", const=True, default=False,
        help="Check for new mangas"
    )
    parser.add_argument(
        "--chapter", nargs="?", type=int, default=None,
        help="Selected chapter"
    )
    parser.add_argument(
        "-r", "--read", nargs="?", const="", default=None,
        help="Mark all as read / specific"
    )
    parser.add_argument(
        "--sync", action="store_true",
        help="Sync with database"
    )
    parser.add_argument(
        "--setup", action="store_true",
        help="Setup database database"
    )
    parser.add_argument(
        "--add", nargs="?", default=None,
        help="Add manga"
    )

    return parser


def format_mangas(mangas):
    """

    :param mangas:
    :type mangas: list[idefix.model.Manga]
    :return:
    :rtype: str | unicode
    """
    if not mangas:
        return ""

    res = []

    for m in mangas:
        url_parts = ""
        if m.urls:
            url_parts = ": " + ",".join([urlparse(u).netloc for u in m.urls])
        int_chap = m.chapter
        if int_chap:
            try:
                int_chap = int(m.chapter)
            except:
                int_chap = m.chapter
        if int_chap != m.chapter:
            int_chap = m.chapter
        res.append("{} ({}){}".format(
            m.name, int_chap, url_parts
        ))
    return "\n".join(res)


def main():
    logging.config.dictConfig(default_logging_config)
    logging.getLogger().setLevel(logging.INFO)

    parser = setup_parser()
    args = parser.parse_args()

    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)
    instance = IDFXManga({
        'settings_file': args.settings,
        'path_prefix': args.pre_path
    })

    # instance.start(blocking=False)
    user, mangas = instance.load_manga_file(path=args.manga_file)
    new = None
    dirty = False
    if args.add:
        n = Manga(name=args.add)
        found = [m for m in mangas if n.name.lower() == m.name.lower()]
        if found:
            logger.info("Manga already added ({})".format(found[0].name))
        else:
            n.updated = datetime.datetime.utcnow()
            n.created = datetime.datetime.utcnow()
            mangas.append(n)
            dirty = True
    if args.check:
        new = instance.check(mangas, instance.create_index())
        # logger.debug(new)
        if new:
            logger.info("\n" + format_mangas(new))
        else:
            logger.info("No updates")
    if args.read is not None:
        if new is None:
            new = instance.check(mangas, instance.create_index())
        if new:
            r = args.read.lower()
            for m in mangas:
                if r and not m.name.lower().startswith(r):
                    continue
                for n in new:
                    if not((n.name or m.name)
                           and n.name.lower() == m.name.lower()):
                        if not((n.uuid or m.uuid) and n.uuid == m.uuid):
                            continue
                    m.chapter = n.chapter
                    m.updated = n.updated
                    m.urls = n.urls
                    dirty = True
                    logger.info("Read " + format_mangas([m]))
        else:
            logger.info("Nothing read")
    if args.setup:
        if not instance.dao._is_running:
            instance.dao.start(False)
        instance.dao.setup()
    if args.sync:
        if not instance.dao._is_running:
            instance.dao.start(False)
        db_mangas = instance.dao.read_get(user)
        db_dict = {db.name.lower(): db for db in db_mangas}
        file_dict ={m.name.lower(): m for m in mangas}
        db_set = set(db_dict.keys())
        file_set = set(file_dict.keys())
        missing_db = [file_dict[key] for key in (file_set - db_set)]
        missing_file = [db_dict[key] for key in (db_set - file_set)]
        upd_db = []

        for key in (db_set & file_set):
            d = db_dict[key]
            m = file_dict[key]
            if not m.uuid:
                logger.debug("{} -> {}".format(d.uuid, m.uuid))
                m.uuid = d.uuid
                dirty = True
            if not m.created:
                logger.debug("{} -> {}".format(d.created, m.created))
                m.created = d.created
                dirty = True
            if not m.updated:
                logger.debug("{} -> {}".format(d.updated, m.updated))
                m.updated = d.updated
                dirty = True
            if not m.name:
                logger.debug("{} -> {}".format(d.name, m.name))
                m.name = d.name
                dirty = True
            if d.updated < m.updated:
                # m newer
                logger.debug("{} -> {}".format(m, d))
                d.name = m.name
                d.chapter = m.chapter
                d.updated = m.updated
                upd_db.append(d)
            elif d.updated > m.updated:
                # d newer
                logger.debug("{} -> {}".format(d, m))
                m.name = d.name
                m.chapter = d.chapter
                m.updated = d.updated
                dirty = True
            elif d.chapter != m.chapter:
                if m.chapter is None or m.chapter < d.chapter:
                    logger.debug("{} -> {}".format(d.chapter, m.chapter))
                    m.chapter = d.chapter
                    dirty = True
                elif d.chapter is None or d.chapter < m.chapter:
                    logger.debug("{} -> {}".format(m.chapter, d.chapter))
                    d.chapter = m.chapter
                    upd_db.append(d)
        if missing_db:
            logger.debug("Missing from db:\n" + format_mangas(missing_db))
            for d in missing_db:
                if not d.uuid:
                    dirty = True
                instance.create_manga(user, d)
        if upd_db:
            logger.debug("Update in db:\n" + format_mangas(upd_db))
            for d in upd_db:
                instance.dao.read_update(user, d)
        if missing_file:
            logger.debug("Missing from file:\n" + format_mangas(missing_file))
            for m in missing_file:
                mangas.append(m)
                dirty = True
        instance.dao.stop()

    if dirty:
        logger.debug("Dirty")
        instance.save_manga_file(user, mangas, readable=True)



# -*- coding: UTF-8 -*-
__author__ = "d01"
__copyright__ = "Copyright (C) 2015, Florian JUNG"
__license__ = "All rights reserved"
__version__ = "0.1.0a0"
__date__ = "2015-03-13"
# Created: 2015-03-13 12:11



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__copyright__ = "Copyright (C) 2015-17, Florian JUNG"
__license__ = "All rights reserved"
__version__ = "0.2.1"
__date__ = "2017-12-07"
# Created: 2015-03-13 12:34

import datetime
import uuid

import pymysql
from flotils import Loadable, StartStopable, StartException

from ..model import Manga, User
from ..errors import DAOException, AlreadyExistsException, ValueException


class SqlConnector(Loadable, StartStopable):
    """ Connect to mysql database """

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        super(SqlConnector, self).__init__(settings)
        self._server = settings['server']
        self._default_db = settings['database']
        self._user = settings['user']
        self._pw = settings['password']
        self.connection = None
        """ Current connection
            :type : pymysql.Connection """
        self.cursor = None
        """ Current cursor
            :type : None | pymysql.cursors.Cursor """
        self._dbs = {}
        """ :type : dict[str | unicode, (pymysql.Connection, pymysql.cursors.Cursor)] """
        self.database = None
        """ Current database
            :type : None | str | unicode """

    def _connect(self, db=None, encoding="utf8"):
        """
        Connect to database

        :param db: Database to connect to
        :type db: str | unicode
        :param encoding: Database encoding
        :type db: str | unicode
        :rtype: None
        :raises DAOException: Failed to connect
        """
        if not db:
            db = self._default_db
        try:
            con = pymysql.connect(
                host=self._server,
                user=self._user,
                passwd=self._pw,
                db=db,
                charset=encoding
            )
            cur = con.cursor()
            cur.execute("SET time_zone= '+00:00'")
        except Exception as e:
            self.exception("Failed to connect")
            raise DAOException(e)
        self._dbs[db] = (con, cur)
        self.database = db
        self.connection = con
        self.cursor = cur

    def _close(self, db=None):
        """
        Commit and close database connection

        :param db: Database to connect to (default: None)
                    None means close all
        :type db: None | str | unicode
        :rtype: None
        """
        if db:
            dbs = [db]
        else:
            dbs = self._dbs.keys()

        for db in dbs:
            con, cur = self._dbs[db]

            try:
                if cur:
                    cur.close()
                if con:
                    con.commit()
                    con.close()
            except:
                self.exception("Failed to close")
            finally:
                del self._dbs[db]

                if db == self.database:
                    self.connection = None
                    self.cursor = None
                    self.database = None

    def switch_database(self, db):
        """
        Switch to database

        :param db: Database to switch to
        :type db: str | unicode
        :rtype: None
        :raises DAOException: Failed to connect
        """
        if db is None:
            db = self._default_db
        if db == self.database:
            return

        if db in self._dbs:
            # Update old values
            self._dbs[self.database] = (self.connection, self.cursor)
            # Get current
            self.connection, self.cursor = self._dbs[db]
            self.database = db
            return
        else:
            self._connect(db)

    def execute(self, cmd, args=None, db=None):
        """
        Execute a sql command

        :param cmd: Sql command to execute
        :type cmd: str | unicode
        :param args: Arguments to add (default: None)
        :type args: None | tuple[str | unicode]
        :param db: Database to execute on (default: None)
                    None means on current
        :type db:  str | unicode | None
        :return: result of .execute()
        :raises DAOException: Failed to connect
        """
        if db is not None and db != self.database:
            self.switch_database(db)

        try:
            if args:
                return self.cursor.execute(cmd, args)
            else:
                return self.cursor.execute(cmd)
        except Exception as e:
            raise DAOException(e)

    def fetchall(self, db=None):
        """
        Get all rows for database

        :param db: Database to get from (default: None)
                    None means current
        :type db: None | str | unicode
        :return: Result of .fetchall()
        :rtype: collections.iterable
        :raises DAOException: Failed to connect
        """
        if db is not None and db != self.database:
            self.switch_database(db)

        try:
            return self.cursor.fetchall()
        except Exception as e:
            raise DAOException(e)

    def commit(self, db=None):
        if db is not None and db != self.database:
            self.switch_database(db)

        try:
            return self.connection.commit()
        except Exception as e:
            raise DAOException(e)

    def rollback(self, db=None):
        if db is not None and db != self.database:
            self.switch_database(db)

        try:
            return self.connection.rollback()
        except Exception as e:
            raise DAOException(e)

    def change_encoding(self, charset, db=None):
        self.execute(
            """ALTER DATABASE python CHARACTER SET '%s'""",
            args=charset,
            db=db
        )
        return self.fetchall()

    def get_encodings(self, db=None):
        self.execute("""SHOW variables LIKE '%character_set%'""", db=db)
        return self.fetchall()

    def setup(self):
        if not self.connection:
            self._connect()

        query = "CREATE FUNCTION IF NOT EXISTS UuidToBin(_uuid BINARY(36))" \
                "  RETURNS BINARY(16)" \
                "  LANGUAGE SQL  DETERMINISTIC  CONTAINS SQL  " \
                "SQL SECURITY INVOKER" \
                " RETURN" \
                "  UNHEX(CONCAT(" \
                "    SUBSTR(_uuid, 15, 4)," \
                "    SUBSTR(_uuid, 10, 4)," \
                "    SUBSTR(_uuid,  1, 8)," \
                "    SUBSTR(_uuid, 20, 4)," \
                "    SUBSTR(_uuid, 25)" \
                "  ));" \
                "\n\n" \
                "CREATE FUNCTION IF NOT EXISTS UuidFromBin(_bin BINARY(16))" \
                "  RETURNS BINARY(36)" \
                "  LANGUAGE SQL  DETERMINISTIC  CONTAINS SQL  " \
                "SQL SECURITY INVOKER" \
                " RETURN" \
                "  LCASE(CONCAT_WS('-'," \
                "      HEX(SUBSTR(_bin,  5, 4))," \
                "      HEX(SUBSTR(_bin,  3, 2))," \
                "      HEX(SUBSTR(_bin,  1, 2))," \
                "      HEX(SUBSTR(_bin,  9, 2))," \
                "      HEX(SUBSTR(_bin, 11))" \
                "  ));"

        try:
            res = self.execute(query)
            if res:
                self.info("Created functions")
        except:
            self.exception("Failed to create functions")
            raise DAOException('Functions failed')

        query = "CREATE TABLE IF NOT EXISTS mangas (" \
                " uuid BINARY(16) NOT NULL UNIQUE PRIMARY KEY," \
                " created DATETIME NOT NULL," \
                " updated DATETIME NOT NULL," \
                " name VARCHAR(255) NOT NULL UNIQUE," \
                " latest_chapter DECIMAL(8,2) " \
                ") DEFAULT CHARACTER SET utf8 COLLATE utf8_general_ci;"

        try:
            res = self.execute(query)
            if res:
                self.info("Created table mangas")
        except:
            self.exception("Failed to create table mangas")
            raise DAOException('Table create failed')

        query = "CREATE TABLE IF NOT EXISTS users (" \
                " uuid BINARY(16) NOT NULL UNIQUE PRIMARY KEY," \
                " created DATETIME NOT NULL," \
                " updated DATETIME NOT NULL," \
                " firstname VARCHAR(255) NOT NULL," \
                " lastname VARCHAR(255) NOT NULL," \
                " role INTEGER NOT NULL," \
                " CONSTRAINT users_name_uq UNIQUE (lastname, firstname)" \
                ") DEFAULT CHARACTER SET utf8 COLLATE utf8_general_ci;"

        try:
            res = self.execute(query)
            if res:
                self.info("Created table users")
        except:
            self.exception("Failed to create table users")
            raise DAOException('Table create failed')

        query = "CREATE TABLE IF NOT EXISTS mangas_read (" \
            " user_uuid BINARY(16) NOT NULL," \
            " manga_uuid BINARY(16) NOT NULL," \
            " created DATETIME NOT NULL," \
            " updated DATETIME NOT NULL," \
            " chapter DECIMAL(8,2)," \
            " CONSTRAINT mangas_read_pk PRIMARY KEY (user_uuid, manga_uuid)," \
            " CONSTRAINT mangas_read_user_fk FOREIGN KEY (user_uuid)" \
            "   REFERENCES users (uuid) ON DELETE CASCADE," \
            " CONSTRAINT mangas_read_manga_fk FOREIGN KEY (manga_uuid)" \
            "   REFERENCES mangas (uuid) ON DELETE CASCADE," \
            " CONSTRAINT mangas_read_pk_uq UNIQUE (user_uuid, manga_uuid)" \
            ") DEFAULT CHARACTER SET utf8 COLLATE utf8_general_ci;"

        try:
            res = self.execute(query)
            if res:
                self.info("Created table mangas_read")
        except:
            self.exception("Failed to create table mangas_read")
            raise DAOException('Table create failed')

    def manga_create(self, manga):
        """
        New manga

        :param manga: Manga to create
        :type manga: idefix.model.Manga
        :return: Affected entries
        :rtype: int
        :raises DAOException: Failure
        """
        if not manga:
            raise ValueException("Invalid manga")
        if not manga.created:
            manga.created = datetime.datetime.utcnow()
        if not manga.updated:
            manga.updated = manga.created
        if not manga.uuid:
            manga.uuid = "{}".format(uuid.uuid4())
        try:
            affected = self.execute(
                "INSERT INTO mangas ("
                "  uuid,created,updated,name,latest_chapter"
                ") VALUES "
                "(UuidToBin(%s),%s,%s,%s,%s);",
                (
                    manga.uuid, manga.created, manga.updated,
                    manga.name, manga.latest_chapter
                )
            )
            # self.debug(affected)
            return affected
        except DAOException as e:
            if e.args[0] and e.args[0].args and e.args[0].args[0] == 1062:
                # Double entry
                raise AlreadyExistsException(e.args[0].args[1])
            raise
        except Exception as e:
            raise DAOException(e)

    def manga_update(self, manga):
        """
        Update manga

        :param manga: Manga to update
        :type manga: idefix.model.Manga
        :return: Affected entries
        :rtype: int
        :raises DAOException: Failure
        """
        if not manga or not manga.uuid:
            raise ValueException("Invalid manga")
        if not manga.updated:
            manga.updated = datetime.datetime.utcnow()
        query = "UPDATE mangas SET updated=%s"
        args = (manga.updated,)
        if manga.name is not None:
            query += ",name=%s"
            args += (manga.name,)
        if manga.latest_chapter is not None:
            query += ",latest_chapter=%s"
            args += (manga.latest_chapter,)
        try:
            affected = self.execute(
                query + " WHERE UuidFromBin(uuid)=%s",
                args + (manga.uuid,)
            )
            # self.debug(affected)
            return affected
        except DAOException as e:
            if e.args[0] and e.args[0].args and e.args[0].args[0] == 1062:
                # Double entry
                raise AlreadyExistsException(e.args[0].args[1])
            raise
        except Exception as e:
            raise DAOException(e)

    def manga_delete(self, manga):
        """
        Delete manga

        :param manga: Manga to delete
        :type manga: idefix.model.Manga
        :return: Affected entries
        :rtype: int
        :raises DAOException: Failure
        """
        if not manga or not manga.uuid:
            raise ValueException("Invalid manga")
        try:
            affected = self.execute(
                "DELETE FROM mangas WHERE UuidFromBin(uuid)=%s",
                (manga.uuid,)
            )
            # self.debug(affected)
            return affected
        except DAOException:
            raise
        except Exception as e:
            raise DAOException(e)

    def manga_get(self, manga=None, use_like=False):
        """
        Get manga

        :param manga: Manga to get or None for all (default: None)
        :type manga: None | idefix.model.Manga
        :param use_like: Use LIKE for comparison (default: False)
        :type use_like: bool
        :return: Found mangas
        :rtype: list[idefix.model.Manga]
        :raises DAOException: Failure
        """
        args = ()
        where_query = ""
        where_query_parts = []
        if manga:
            if manga.name:
                op = "LIKE" if use_like else "="
                where_query_parts.append(("name", op, manga.name))
            if manga.uuid:
                # Unique, other values not necessary
                where_query_parts = [("UuidFromBin(uuid)", "=", manga.uuid)]

        if where_query_parts:
            where_query = " WHERE " + " AND ".join(
                ["{} {} %s".format(t[0], t[1]) for t in where_query_parts]
            )
            args = tuple([t[2] for t in where_query_parts])
        mangas = []
        try:
            affected = self.execute(
                "SELECT"
                " UuidFromBin(uuid),created,updated,name,latest_chapter "
                "FROM mangas" + where_query
                , args)
            # self.debug(affected)
            if affected:
                res = self.fetchall()
                for t in res:
                    mangas.append(Manga(
                        uuid=t[0], name=t[3]
                    ))
                    mangas[-1].created = t[1]
                    mangas[-1].updated = t[2]
                    mangas[-1].latest_chapter = float(t[4])
        except DAOException:
            raise
        except Exception as e:
            raise DAOException(e)
        return mangas

    def user_create(self, user):
        """
        New user

        :param user: User to create
        :type user: idefix.model.User
        :return: Affected entries
        :rtype: int
        :raises DAOException: Failure
        """
        if not user:
            raise ValueException("Invalid user")
        if not user.created:
            user.created = datetime.datetime.utcnow()
        if not user.updated:
            user.updated = user.created
        if not user.uuid:
            user.uuid = "{}".format(uuid.uuid4())
        try:
            affected = self.execute(
                "INSERT INTO users ("
                "  uuid,created,updated,firstname,lastname,role"
                ") VALUES "
                "(UuidToBin(%s),%s,%s,%s,%s,%s);",
                (
                    user.uuid, user.created, user.updated,
                    user.firstname, user.lastname, user.role
                )
            )
            # self.debug(affected)
            return affected
        except DAOException as e:
            if e.args[0] and e.args[0].args and e.args[0].args[0] == 1062:
                # Double entry
                raise AlreadyExistsException(e.args[0].args[1])
            raise
        except Exception as e:
            raise DAOException(e)

    def user_update(self, user):
        """
        Update user

        :param user: User to update
        :type user: idefix.model.User
        :return: Affected entries
        :rtype: int
        :raises DAOException: Failure
        """
        if not user or not user.uuid:
            raise ValueException("Invalid user")
        if not user.updated:
            user.updated = datetime.datetime.utcnow()
        query = "UPDATE users SET updated=%s"
        args = (user.updated,)
        if user.firstname is not None:
            query += ",firstname=%s"
            args += (user.firstname,)
        if user.lastname is not None:
            query += ",lastname=%s"
            args += (user.lastname,)
        try:
            affected = self.execute(
                query + " WHERE UuidFromBin(uuid)=%s",
                args + (user.uuid,)
            )
            # self.debug(affected)
            return affected
        except DAOException as e:
            if e.args[0] and e.args[0].args and e.args[0].args[0] == 1062:
                # Double entry
                raise AlreadyExistsException(e.args[0].args[1])
            raise
        except Exception as e:
            raise DAOException(e)

    def user_delete(self, user):
        """
        Delete user

        :param user: User to delete
        :type user: idefix.model.User
        :return: Affected entries
        :rtype: int
        :raises DAOException: Failure
        """
        if not user or not user.uuid:
            raise ValueException("Invalid user")
        try:
            affected = self.execute(
                "DELETE FROM users WHERE UuidFromBin(uuid)=%s",
                (user.uuid,)
            )
            # self.debug(affected)
            return affected
        except DAOException:
            raise
        except Exception as e:
            raise DAOException(e)

    def user_get(self, user=None, use_like=False):
        """
        Find user

        :param user: User to get or None for all (default: None)
        :type user: None | idefix.model.User
        :param use_like: Use LIKE for comparison (default: False)
        :type use_like: bool
        :return: Found users
        :rtype: list[idefix.model.User]
        :raises DAOException: Failure
        """
        args = ()
        where_query = ""
        where_query_parts = []
        if user:
            if user.firstname:
                op = "LIKE" if use_like else "="
                where_query_parts.append(
                    ("firstname", op, user.firstname.lower())
                )
            if user.lastname:
                op = "LIKE" if use_like else "="
                where_query_parts.append(
                    ("lastname", op, user.lastname.lower())
                )
            if user.uuid:
                # Unique, other values not necessary
                where_query_parts = [("UuidFromBin(uuid)", "=", user.uuid)]

        if where_query_parts:
            where_query = " WHERE " + " AND ".join(
                ["{} {} %s".format(t[0], t[1]) for t in where_query_parts]
            )
            args = tuple([t[2] for t in where_query_parts])
        users = []
        try:
            affected = self.execute(
                "SELECT"
                " UuidFromBin(uuid),created,updated,firstname,lastname,role "
                "FROM users" + where_query
                , args)
            # self.debug(affected)
            if affected:
                res = self.fetchall()
                for t in res:
                    users.append(User(
                        uuid=t[0], firstname=t[3], lastname=t[4]
                    ))
                    users[-1].created = t[1]
                    users[-1].updated = t[2]
                    users[-1].role = t[5]
        except DAOException:
            raise
        except Exception as e:
            raise DAOException(e)
        return users

    def read_create(self, user, manga):
        """
        New user read manga

        :param user: User to create
        :type user: idefix.model.User
        :param manga: Manga to create
        :type manga: idefix.model.Manga
        :return: Affected entries
        :rtype: int
        :raises DAOException: Failure
        """
        if not user or not user.uuid:
            raise ValueException("Invalid user")
        if not manga or not manga.uuid:
            raise ValueException("Invalid manga")
        chapter = 0
        if manga.chapter:
            chapter = manga.chapter
        now = datetime.datetime.utcnow()
        if not manga.created:
            manga.created = now
        if not manga.updated:
            manga.updated = now
        try:
            affected = self.execute(
                "INSERT INTO mangas_read ("
                "  user_uuid,manga_uuid,created,updated,chapter"
                ") VALUES "
                "(UuidToBin(%s),UuidToBin(%s),%s,%s,%s);",
                (user.uuid, manga.uuid, manga.created, manga.updated, chapter)
            )
            # self.debug(affected)
            return affected
        except DAOException as e:
            if e.args[0] and e.args[0].args and e.args[0].args[0] == 1062:
                # Double entry
                raise AlreadyExistsException(e.args[0].args[1])
            raise
        except Exception as e:
            raise DAOException(e)

    def read_update(self, user, manga):
        """
        Update user read manga

        :param user: User to update
        :type user: idefix.model.User
        :param manga: Manga to update
        :type manga: idefix.model.Manga
        :return: Affected entries
        :rtype: int
        :raises DAOException: Failure
        """
        if not user or not user.uuid:
            raise ValueException("Invalid user")
        if not manga or not manga.uuid or manga.chapter is None:
            raise ValueException("Invalid manga")
        now = datetime.datetime.utcnow()
        try:
            affected = self.execute(
                "UPDATE mangas_read SET updated=%s,chapter=%s"
                " WHERE "
                "UuidFromBin(user_uuid)=%s AND UuidFromBin(manga_uuid)=%s",
                (now, manga.chapter, user.uuid, manga.uuid)
            )
            # self.debug(affected)
            return affected
        except DAOException as e:
            if e.args[0] and e.args[0].args and e.args[0].args[0] == 1062:
                # Double entry
                raise AlreadyExistsException(e.args[0].args[1])
            raise
        except Exception as e:
            raise DAOException(e)

    def read_delete(self, user, manga):
        """
        Delete user read manga

        :param user: User to delete
        :type user: idefix.model.User
        :param manga: Manga to delete
        :type manga: idefix.model.Manga
        :return: Affected entries
        :rtype: int
        :raises DAOException: Failure
        """
        if not user or not user.uuid:
            raise ValueException("Invalid user")
        if not manga or not manga.uuid:
            raise ValueException("Invalid manga")
        try:
            affected = self.execute(
                "DELETE FROM mangas_read"
                " WHERE "
                "UuidFromBin(user_uuid)=%s AND UuidFromBin(manga_uuid)=%s",
                (user.uuid, manga.uuid)
            )
            # self.debug(affected)
            return affected
        except DAOException:
            raise
        except Exception as e:
            raise DAOException(e)

    def read_get(self, user, manga=None, use_like=False):
        """
        Find user read manga

        :param user: User to find for
        :type user: idefix.model.User
        :param manga: Manga being read to get or None for all (default: None)
        :type manga: None | idefix.model.Manga
        :param use_like: Use LIKE for comparison (default: False)
        :type use_like: bool
        :return: Found mangas
        :rtype: list[idefix.model.Manga]
        :raises DAOException: Failure
        """
        if not user or not user.uuid:
            raise ValueException("Invalid user")
        args = (user.uuid,)
        where_query = " WHERE UuidFromBin(user_uuid)=%s"
        where_query_parts = []
        if manga:
            if manga.name:
                op = "LIKE" if use_like else "="
                where_query_parts.append(("m.name", op, manga.name))
            if manga.chapter:
                where_query_parts.append(("mr.chapter", "=", manga.chapter))
            if manga.uuid:
                # Unique, other values not necessary
                where_query_parts = [(
                    "UuidFromBin(mr.manga_uuid)", "=", manga.uuid
                )]

        if where_query_parts:
            where_query += " AND " + " AND ".join(
                ["{} {} %s".format(t[0], t[1]) for t in where_query_parts]
            )
            args += tuple([t[2] for t in where_query_parts])
        mangas = []
        try:
            affected = self.execute(
                "SELECT"
                " UuidFromBin(m.uuid),m.name,m.latest_chapter,"
                "mr.created,mr.updated,mr.chapter "
                "FROM mangas_read mr JOIN mangas m ON mr.manga_uuid = m.uuid"
                + where_query
                , args)
            # self.debug(affected)
            if affected:
                res = self.fetchall()
                for t in res:
                    mangas.append(Manga(
                        uuid=t[0], name=t[1], chapter=float(t[5])
                    ))
                    mangas[-1].created = t[3]
                    mangas[-1].updated = t[4]
                    mangas[-1].latest_chapter = float(t[2])
        except DAOException:
            raise
        except Exception as e:
            raise DAOException(e)
        return mangas

    def read_get_index(self):
        """
        Get mangas linked to users reading them

        :return: Manga - user/chapter correlation
        :rtype: list[(idefix.model.Manga, list[(str | unicode, int)]]
        """
        mangas = []
        try:
            affected = self.execute(
                "SELECT"
                " UuidFromBin(m.uuid),m.name,"
                " mr.chapter,UuidFromBin(mr.user_uuid) "
                "FROM mangas_read mr JOIN mangas m ON mr.manga_uuid = m.uuid "
                "ORDER BY m.uuid"
            )
            if affected:
                res = self.fetchall()
                for t in res:
                    if not mangas or mangas[-1][0].uuid != t[0]:
                        mangas.append((
                            Manga(uuid=t[0], name=t[1]), []
                        ))
                    mangas[-1][1].append((t[3], float(t[2])))
        except DAOException:
            raise
        except Exception as e:
            raise DAOException(e)
        # sanity check
        mids = [m.uuid for m, _ in mangas]
        smids = set(mids)
        if len(mids) != len(smids):
            self.error("Got double entries {}-{}".format(len(mids), len(smids)))
        return mangas

    def start(self, blocking=False):
        self.debug("()")
        try:
            self._connect()
        except:
            # self.exception("Failed to connect")
            raise StartException("Connecting failed")
        super(SqlConnector, self).start(blocking)

    def stop(self):
        self.debug("()")
        try:
            self._close()
        except:
            self.exception("Failed to close")
        super(SqlConnector, self).stop()



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-11-30"
# Created: 2017-11-27 21:09


class IDFXException(Exception):
    """ IDefix base exception """
    pass


class FileException(IDFXException):
    """ File related exception (IOError?) """
    pass


class NoDAOException(IDFXException):
    """ No DAO available """
    pass

class DAOException(IDFXException):
    """ Base DAO exception """
    pass


class ValueException(DAOException, ValueError):
    """ DAO value error """
    pass


class AlreadyExistsException(DAOException):
    """ Entry already exits """
    pass



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# from __future__ import unicode_literals

__author__ = "d01 <Florian Jung>"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2015-17, Florian JUNG"
__license__ = "MIT"
__version__ = "0.2.1"
__date__ = "2016-05-12"
# Created: ?

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
import sys
import os
import re


if sys.argv[-1] == "build":
    os.system("python setup.py clean sdist bdist bdist_egg bdist_wheel")


def get_version():
    """
    Parse the version information from the init file
    """
    version_file = os.path.join("idfx", "__init__.py")
    initfile_lines = open(version_file, "rt").readlines()
    version_reg = r"^__version__ = ['\"]([^'\"]*)['\"]"
    for line in initfile_lines:
        mo = re.search(version_reg, line, re.M)
        if mo:
            return mo.group(1)
    raise RuntimeError(
        u"Unable to find version string in {}".format(version_file)
    )


def get_file(path):
    with open(path, "r") as f:
        return f.read()


def split_external_requirements(requirements):
    external = []
    # External dependencies
    pypi = []
    # Dependencies on pypi

    for req in requirements:
        if req.startswith("-e git+"):
            # External git link
            external.append(req.lstrip("-e git+"))
        else:
            pypi.append(req)
    return pypi, external


version = get_version()
readme = get_file("README.rst")
history = get_file("HISTORY.rst")
pypi, external = split_external_requirements(
    get_file("requirements.txt").split("\n")
)

assert version is not None
assert readme is not None
assert history is not None
assert pypi is not None
assert external is not None

setup(
    name="idfx",
    version=version,
    description="Manga chapter checker",
    long_description=readme + "\n\n" + history,
    author="the01",
    author_email="jungflor@gmail.com",
    url="https://github.com/the01/python-idefix",
    packages=[
        "idfx",
        "idfx.dao"
    ],
    install_requires=pypi,
    dependency_links=external,
    entry_points={
        'console_scripts': [
            "idefix=idfx.cli:main",
        ]
    },
    license="MIT License",
    keywords="idefix manga",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
    ]
)



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2013-17, Florian JUNG"
__license__ = "MIT"
__version__ = "0.2.1"
__date__ = "2017-07-15"
# Created: 2013-08-04 24:00

import threading
import multiprocessing
import multiprocessing.pool
import os
import datetime

from flotils.loadable import Loadable, save_file, load_file
from floscraper.webscraper import WebScraper, WEBConnectException
from requests.compat import urljoin
from dateutil.tz import tzutc

from .model import Manga, User
from .dao.mysql import SqlConnector
from .errors import AlreadyExistsException, DAOException, ValueException,\
    NoDAOException, IDFXException


def to_utc(dt):
    dt = dt.astimezone(tzutc())
    return dt.replace(tzinfo=None)


class IDFXManga(Loadable):
    """ Class for retrieving manga info """

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        super(IDFXManga, self).__init__(settings)

        self._read_path = self.join_path_prefix(
            settings.get('manga_path', None)
        )

        self.dao = None
        """ :type : None | idefix.dao.mysql.SqlConnector """
        if 'dao' in settings:
            self.dao = SqlConnector(settings['dao'])
        else:
            self.warning("No DAO")
        self.scrapers = settings.get('scrapers', None)
        self._threading_mode = None
        self.pool = None
        self.pool_size = settings.get('pool_size', multiprocessing.cpu_count())
        self.threading_mode = settings.get('threading_mode', None)

    @property
    def threading_mode(self):
        return self._threading_mode

    @threading_mode.setter
    def threading_mode(self, value):
        if self._threading_mode == value:
            return
        if self.pool:
            # TODO: stop
            self.pool = None
        if value == "pool":
            self.pool = multiprocessing.pool.ThreadPool(
                processes=self.pool_size
            )
        self._threading_mode = value

    def load_manga_file(self, user=None, path=None):
        """
        Load manga list from file
        Either using exact path or user and base path

        :param user: User to load for
        :type user: None | idefix.model.User
        :param path: Base path / exact path to load from
        :type path: None | str | unicode
        :return: Loaded user, mangas
        :rtype: idefix.model.Idefix, list[idefix.model.Manga]
        """
        if not path:
            path = self._read_path
        if not path:
            path = ""
        user_id = None
        if user:
            user_id = user.uuid
            if not user_id:
                user_id = "{}_{}".format(user.lastname, user.firstname)

        if not path or not os.path.isfile(path):
            path = os.path.join(path, "idfx_manga_{}.yaml".format(user_id))

        if not os.path.isfile(path):
            self.error("File not found {}".format(path))
            raise IOError("File not found {}".format(path))

        try:
            loaded = load_file(path)
        except IOError:
            raise
        except Exception as e:
            self.exception("Failed to load manga file")
            raise IOError(e)
        if "user" not in loaded:
            raise ValueException("User missing")
        if "mangas" not in loaded:
            raise ValueException("Mangas missing")
        mangas = []
        for mdict in loaded['mangas']:
            m = Manga.from_dict(mdict)
            """ :type : idefix.model.Manga """
            if m.updated:
                m.updated = to_utc(m.updated)
            if m.created:
                m.created = to_utc(m.created)
            mangas.append(m)
        return User.from_dict(loaded['user']), mangas

    def save_manga_file(self, user, mangas, path=None, readable=False):
        """
        Save manga list to file

        :param user: User to save for
        :type user: idefix.model.User
        :param mangas: List of mangas of user
        :type mangas: list[idefix.model.Manga]
        :param path: Base path to save to or _read_path (default: None)
        :type path: None | str | unicode
        :param readable: Format content to be human readable (default: False)
        :type readable: bool
        :rtype: None
        """
        if not path:
            path = self._read_path
        if not path:
            path = ""
        user_id = user.uuid
        if not user_id:
            user_id = "{}_{}".format(user.lastname, user.firstname)
        temp_path = os.path.join(path, "idfx_manga_{}.json".format(user_id))

        if os.path.exists(os.path.basename(temp_path)):
            path = temp_path

        save_file(path, {
            'user': user.to_dict(),
            'mangas': [m.to_dict() for m in sorted(mangas, key=lambda a: a.name)]
        }, readable)

    def create_manga(self, user, manga):
        """
        New manga for user

        :param user: User to create for
        :type user: idefix.model.User
        :param manga: Manga to add to user
        :type manga: idefix.model.Manga
        :return: If multiple found, list of mangas to select or affected rows
        :rtype: list[idefix.model.Manga] | int
        :raises IDFXException: Failure
        :raises ValueException: Failure
        """
        self.debug("Adding manga '{}' for user '{}'".format(manga, user))
        if not user or not user.uuid:
            raise ValueException("Invalid user")
        if not manga or not (manga.uuid or manga.name):
            raise ValueException("Invalid manga")
        if not manga.uuid:
            # Is unknown manga?
            mangas = self.dao.manga_get(Manga(name=manga.name))
            if not mangas:
                # Create new manga
                self.info("Creating new manga: {}".format(manga.name))
                if self.dao.manga_create(manga) == 0:
                    raise DAOException("Manga not created")
                mangas = self.dao.manga_get(manga)
            if not mangas:
                raise ValueException("Manga not found")
            if len(mangas) > 1:
                return mangas
            manga.uuid = mangas[0].uuid
            manga.created = mangas[0].created
            manga.updated = mangas[0].updated
        res = self.dao.read_create(user, manga)
        if not res:
            raise DAOException("Manga read not created")
        self.dao.commit()
        return res

    def _do_scrap(self, scraper):
        """
        Perform thread-safe scrap step and update _upd

        :param scraper: Used scraper
        :type scraper: WebScraper
        :rtype: None
        """
        try:
            res = scraper.scrap()
        except WEBConnectException as e:
            self.error("{}".format(e))
            return None
        except:
            self.exception("Failed to scrap")
            return None #raise?

        res.scraped = scraper.shrink(res.scraped)
        url = scraper.url
        url = url.lstrip("https://")
        url = url.lstrip("http://")
        url = url.lstrip("www.")
        url = url.rstrip("/")

        if "chapter" not in res.scraped:
            self.error("Failed to retrieve mangas {}".format(url))
            return None

        mangas = {}
        for a in res.scraped['chapter']:
            number = name = link = ""
            # if value == list -> assumes correct one at index 0
            for val in a:
                if val == "number":
                    # Handle multiple chapters separatly
                    continue
                if isinstance(a[val], list) and len(a[val]) > 0:
                    a[val] = a[val][0]

            if "number" in a:
                numbers = a['number']
                if not isinstance(numbers, list):
                    numbers = [numbers]
                number = -1

                for n in numbers:
                    try:
                        num_candidate = float("{}".format(n).strip())
                        if num_candidate > number:
                            number = num_candidate
                    except:
                        # Not a number -> skip
                        pass
                if number == -1:
                    number = ""
            if not number:
                # No number -> skip
                continue

            if "name" in a:
                name = "{}".format(a['name']).strip()
            if "link" in a:
                link = urljoin(scraper.url, "{}".format(a['link']).strip())

            key = name.lower()
            mangas.setdefault(key, {
                'chapter': 0, 'url': link, 'name': name
            })
            old = mangas[key]['chapter']

            if old < number:
                mangas[key]['chapter'] = number
                mangas[key]['url'] = link
        return mangas

    def create_index(self):
        if not self.scrapers:
            self.error("No scrapers found")
            return None

        for i, scraper in enumerate(self.scrapers):
            if not hasattr(scraper, 'scrap'):
                try:
                    sett = {
                        'path_prefix': self._prePath,
                        'settings_file': scraper
                    }
                    scraper = WebScraper(sett)
                    self.scrapers[i] = scraper
                except:
                    self.exception(
                        "Failed to load scrapper {}".format(scraper)
                    )
                    continue
        if self.threading_mode == "pool" and self.pool:
            self.debug("pool")
            results = self.pool.map(self._do_scrap, self.scrapers)
        elif self.threading_mode == "threaded":
            self.debug("threaded")
            results = []

            def fn(scraper):
                results.append(self._do_scrap(scraper))

            threads = [
                threading.Thread(target=fn, args=(scraper,))
                for scraper in self.scrapers
            ]
            for t in threads:
                t.start()
            for t in threads:
                t.join()
        else:
            # Not threaded
            self.debug("sequential")
            results = [self._do_scrap(scraper) for scraper in self.scrapers]
        index = {}
        """ :type : dict[str | unicode, idefix.model.Manga] """
        for result in results:
            if not result:
                # Skip no mangas/errors
                continue
            for key, value in result.items():
                index.setdefault(
                    key, Manga(name=value['name'], chapter=value['chapter'])
                )
                m = index[key]
                m.updated = datetime.datetime.utcnow()
                if m.chapter and m.chapter == value['chapter']:
                    m.urls.append(value['url'])
                if not m.chapter or m.chapter < value['chapter']:
                    m.urls = [value['url']]
                    m.chapter = value['chapter']
        return index

    def check(self, user_mangas, web_mangas):
        """
        Check for user mangas if there is a new chapter available

        :param user_mangas: Mangas the user is reading
        :type user_mangas: list[idefix.model.Manga]
        :param web_mangas: Mangas that are available on the web
        :type web_mangas: dict[str | unicode, idefix.model.Manga]
        :return: Mangas with updates
        :rtype: list[idefix.model.Manga]
        """
        res = []
        for m in user_mangas:
            if m.name.lower() not in web_mangas:
                # No update for this manga
                continue
            w = web_mangas[m.name.lower()]
            if m.chapter < w.chapter:
                w.uuid = m.uuid
                res.append(w)
        return res

    def check_multiple(self, user_mangas, web_mangas):
        """

        :param user_mangas: Manga -> all users and latest chapters read
        :type user_mangas: list[(idefix.model.Manga, list[(str | unicode, int)]]
        :param web_mangas: Mangas that are available on the web
        :type web_mangas: dict[str | unicode, idefix.model.Manga]
        :return:
        :rtype: dict[str | unicode, list[idefix.model.Manga]]
        """
        res = {}
        for manga, uuidChapter in user_mangas:
            if manga.name.lower() not in web_mangas:
                # No update for this manga
                continue
            w = web_mangas[manga.name.lower()]
            w.uuid = manga.uuid
            for uuid, chapter in uuidChapter:
                if uuid not in res:
                    res[uuid] = []
                if chapter < w.chapter:
                    res[uuid].append(w)
        return res
